#!/usr/local/bin/python

""" Python - tag table for Python (Version 0.6)

    * 0.5->0.6: changed the names of the tags !
                fixed bug in match_str()

    XXX can't handle (lambda ...) and misses not in 'if x is not'

    Copyright (c) 2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
    Copyright (c) 2000-2013, eGenix.com Software GmbH; mailto:info@egenix.com
    See the documentation for further information on copyrights,
    or contact the author. All Rights Reserved.
"""
# engine + constants
from mx.TextTools import *

# helper for correct string matching (with \quotes):

def match_str(s,x,len_text,c):

    while x < len_text and s[x] != c:
        if s[x] == '\\' and s[x+1] == c: 
            x = x + 2
            continue
        x = x + 1
    return x

comment = ('comment',Table,
            ((None,Is,'#'),
             (None,AllNotIn,'\n\r',MatchOk))
          )

whitespace = (None,AllIn,' \t')
opt_whitespace = whitespace + (+1,)

identifier = ('identifier',Table,
              ((None,IsIn,alpha+'_'),
               (None,AllIn,alpha+'_'+number,MatchOk))
             )

string = ('str',Table,
          (# hints
           (None,IsIn,'\"\''),
           (None,Skip,-1),
           # now let's see what we have...
           (None,Word,'"""',+4),
            ('string',NoWord,'"""',+1),
            (None,Word,'"""'),
            (None,Jump,To,MatchOk),
           (None,Word,"'''",+4),
            ('string',NoWord,"'''",+1),
            (None,Word,"'''"),
            (None,Jump,To,MatchOk),
           (None,Is,'"',+4),
            ('string',CallArg,(match_str,'"'),+1),
            (None,Word,'"'),
            (None,Jump,To,MatchOk),
           (None,Is,"'"),
            ('string',CallArg,(match_str,"'"),+1),
            (None,Word,"'"),
            (None,Jump,To,MatchOk))
         )

skw = ["del", "from", "lambda", "return", "and", "elif",
"global", "not", "try", "break", "else", "if", "or", "while",
"except", "import", "pass", "continue", "finally", "in", "print",
"for", "is", "raise"]
keywords = word_in_list(skw)

# note: '=lambda x:...' and '(lambda x:...' are not recognized,
#       yet '= lambda x:...' and '( lambda x:...' are (just like in
#       emacs python-mode) !

keyword = ('kw',Table,
           ((None,AllIn,' \t\n\r'),
            # hints
            (None,IsIn,alpha),
            (None,Skip,-1),
            # one in the list keywords
            ('keyword',Table,keywords,+3),
              (None,IsIn,': \t\n\r'),
              (None,Jump,To,MatchOk),
            # a function declaration
            ('keyword',Word,'def',+12),
              whitespace,
              identifier,
              (None,Is,'('),
              # scan parameters
              ('parameter',AllNotIn,'(),',+2),
              # are there more ?
              (None,Is,',',+1,-1),
              # tuple in param-list ?
              (None,Is,'(',+1,-2),
              # maybe we're done
              (None,Is,')'),
              # to make sure...
              (None,Is,',',+1,-4),
              (None,Is,')',+1),
              # test for correct syntax
              (None,IsIn,': \t\n\r'),
              (None,Jump,To,MatchOk),
            # a class declaration:
            ('keyword',Word,'class'),
              whitespace,
              identifier,
              (None,Is,'(',MatchOk),
              # scan base-classes
              ('baseclass',AllNotIn,'),',+2),
              # are there more ?
              (None,Is,',',+1,-1),
              # we're done
              (None,Is,')'),
              (None,IsIn,': \t\n\r'))
           )

python_script = (comment+(+1,-0),
                 string+(+1,-1),
                 keyword+(+1,-2),
                 # end-of-file ?
                 (None,EOF,Here,+1,MatchOk),
                 # skip uninteresting chars and restart
                 (None,IsIn,any),
                 (None,AllNotIn,'#\'\"_ \n\r\t',-5,-5)
                )

if __name__ == '__main__':

    ### testing

    import sys

    # open file
    text = open(sys.argv[-1]).read()
    print 'read',len(text),'bytes...'
    taglist = []
    t = TextTools._timer()

    # parse file
    t.start()
    result, taglist, next = tag(text,python_script,0,len(text),taglist)
    print 'tag:',t.stop(),'sec.'

    # run against Just's PyFontify if available
    try: 
        import PyFontify # URL: ftp://starship.python.net/pub/crew/just/PyFontify.py
        t.start()
        PyFontify.fontify(text)
        print 'PyFontify:',t.stop(),'sec.'
    except ImportError:
        pass
        
    print 'hit return...'
    raw_input()
    print result
    if result:
        print_tags(text,taglist)



""" Cookie -- Create sendable Netscape Cookies.

    Needs mxDateTime, mxURL and mxTextTools.

    Copyright (c) 2000, Marc-Andre Lemburg; All Rights Reserved.
    See the documentation for further information on copyrights,
    or contact the author. All Rights Reserved.

"""
from mx import DateTime,URL,TextTools

# Version number
__version__ = '1.2'

# Enable debugging output ?
_debug = 0

class Cookie:

    """ Cookie object, cf. Netscape's specification.
    """
    # Default values
    expires = None # temporary cookie
    path = '/'
    domain = ''
    secure = 0

    def __init__(self, name, value, expires=None, 
                 path=None, domain=None, secure=None):

        """ Create a Netscape cookie for name with the given value.

            If expires is given, the cookie will be a temporary cookie
            which expires after a certain amount of time.  expires may
            be given as integer (seconds relative to the current
            time), DateTime instance (absolute date/time) or
            RelativeDateTime instance (relative date/time to current
            time).

            path, domain, secure work according to the Netscape
            specification.
            
        """
        self.name = name
        self.value = value
        if expires is not None:
            # Long living cookie
            if isinstance(expires, DateTime.DateTimeType):
                self.expires = expires.gmtime()
            elif isinstance(expires, DateTime.RelativeDateTime):
                self.expires = DateTime.gmtime() + expires
            else:
                self.expires = DateTime.gmtime() + \
                               expires * DateTime.oneSecond
        if path:
            self.path = path
        if domain:
            self.domain = domain
        if secure:
            self.secure = 1

    def header_content(self,

                       now=DateTime.now,oneSecond=DateTime.oneSecond,
                       urlencode=URL.urlencode):

        # Add things that always have a value
        l = ['%s=%s' % (self.name, urlencode(self.value)),
             'path=%s' % self.path]
        if self.expires is not None:
            l.append('expires=%s' %\
                     self.expires.strftime('%a, %d %b %Y %H:%M:%S GMT'))
        if self.domain:
            l.append('domain=%s' % self.domain)
        if self.secure:
            l.append('secure')
        return ';'.join(l)

    # Alias for b/w compatibility:
    content = header_content

    def set_cookie_header(self):

        return 'Set-Cookie: %s\r\n' % self.header_content()

    # Alias for b/w compatibility:
    http_header = set_cookie_header

    def cookie_header(self,

                      urlencode=URL.urlencode):

        return 'Cookie: %s=%s\r\n' % (self.name, urlencode(self.value))

    def match(self, url, datetime=None):

        """ Return 1/0 depending on whether the Cookie matches
            the given url or not.

            datetime is used to check for expiration in case the
            Cookie is a temporary one. It defaults to the current
            date/time.

        """
        url = URL.URL(url)
        if self.expires is not None:
            if datetime is None:
                datetime = DateTime.now()
            if self.expires < datetime:
                if _debug:
                    print 'expired'
                return 0
        if TextTools.prefix(url.path, (self.path,)) is None:
            if _debug:
                print 'path does not match'
            return 0
        if TextTools.suffix(url.host, (self.domain,)) is None:
            if _debug:
                print 'domain does not match'
            return 0
        return 1

    def __str__(self):

        return 'Set-Cookie: %s' % self.header_content()

    def __repr__(self):

        return '<%s.%s "%s = %s" at 0x%x>' % ( 
            self.__class__.__module__,
            self.__class__.__name__, 
            self.name,
            self.value,
            id(self))

### Set-Cookie header parser

def CookieFromHeader(value,

                     splitat=TextTools.splitat,
                     charsplit=TextTools.charsplit,urldecode=URL.urldecode,
                     lower=TextTools.lower,
                     DateTimeFrom=DateTime.DateTimeFrom):

    """ Parse the value of an Set-Cookie header and return
        a corresponding Cookie instance.

    """
    items = [x.strip() for x in charsplit(value, ';')]
    c = Cookie('no-name', '')
    if not items:
        return c
    c.name, value = splitat(items[0], '=')
    c.value = urldecode(value)
    if len(items) > 1:
        for item in items[1:]:
            key, value = splitat(item,'=')
            key = lower(key)
            if key == 'path':
                c.path = value
            elif key == 'expires':
                try:
                    c.expires = DateTimeFrom(value)
                except ValueError:
                    pass
            elif key == 'domain':
                c.domain = value
            elif key == 'secure':
                c.secure = 1
    return c

# Alias for b/w compatibility
ParserCookie = CookieFromHeader

if __name__ == '__main__':

    c1 = Cookie('a', 'b',
                path='/mysite', expires=20, domain='egenix.com')
    print c1
    s1 = c1.header_content()
    c2 = ParserCookie(s1)
    print c2
    c3 = ParserCookie(s1[:-10])
    print c3
    print
    c4 = Cookie('a', 'b', expires=10)
    print c4.set_cookie_header()
    s1 = c4.header_content()
    c5 = ParserCookie(s1)
    print c5.cookie_header()




""" Tools - A collection of new builtins and misc. helpers for Python

    Copyright (c) 2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
    Copyright (c) 2000-2013, eGenix.com Software GmbH; mailto:info@egenix.com
    See the documentation for further information on copyrights,
    or contact the author. All Rights Reserved.
"""
from Tools import *
from Tools import __version__

# For backward-compatibility: auto-install the add-ons:

### You can control automatic installation with this if-statement.
### Replace the 1 with 0 to disable automatic install.
if 1:
    import NewBuiltins



import sys
from mx.Queue import *

def test():

    s = Queue()
    print repr(s)
    s = Queue()
    for i in range(1000):
          s.push(i)
    while s:
          print s.pop(),
    print

    if 0:
        # which could also be done as:
        s = QueueFromSequence(range(1000))
        while s:
              print s.pop(),

        # or a little different
        s = QueueFromSequence(range(1000))
        print s.as_tuple()
        print s.as_list()
        print

        print 'Pop many.'
        assert s.pop_many(3) == (999, 998, 997)

        print 'Push many.'
        s.push_many(range(100))
        assert s.pop_many(100) == tuple(range(100-1,-1,-1))

        print 'Resize.'
        assert len(s) > 0
        s.resize()

    print 'Clear.'
    s.clear()
    assert len(s) == 0


    if 0:
        print 'Non-zero testing.'
        s.push_many(range(100))
        i = 0
        while s:
            s.pop()
            i = i + 1
        assert i == 100

        # push many + exceptions
        print 'Push many and exceptions.'
        class C:
            def __getitem__(self,i):
                if i < 50:
                    return i + 1
                else:
                    raise IndexError
            def __len__(self):
                return 100
        l = C()

        try:
            s.push_many(l)
        except IndexError:
            pass
        else:
            raise AssertionError,'push_many() does not handle errors correctly'

        assert len(s) == 0

    del s

    # Implementation deleaked up to this line.

    print
    print 'Works.'

if '-m' in sys.argv:
    while 1:
        test()
else:
    test()





#!/usr/local/bin/python -O

""" Tool to show BeeDict data.

"""
import cPickle,sys,os
from BeeDict import *
import CommandLine
from mx import Log

def show(name,keys,all=0,levels=3):

    print 'Records from BeeDict "%s"' % name
    print
    d = BeeDict(name,
                readonly=1)
    if all:
        keys = d.keys()
    for key in keys:
        print_record(d,key,levels)
        print
    d.close()

def print_record(d,key,levels=3):

    print 'Record for key "%s"' % str(key)[:50]
    try:
        data = d[key]
    except:
        print 'Failed to load.'
        return
    try:
        data = cPickle.loads(data)
    except:
        pass
    Log.print_obj(data,indent='  ',levels=levels)

class ShowRecord(CommandLine.Application):

    header = 'Tool for displaying BeeDict records'
    synopsis = '%s [options] dictname keys...'
    options = [CommandLine.ArgumentOption('-d','Display depth',3),
               CommandLine.SwitchOption('-a','Show all records')
               ]

    def check_files(self,files):

        if len(files) < 1:
            self.help('Missing arguments')
            sys.exit(1)

    def main(self):

        show(self.files[0],self.files[1:],self.values['-a'],self.values['-d'])

if __name__ == '__main__':
    ShowRecord()
        



""" ODMG type classes for date/time handling

    These are built on top of the basic DateTime[Delta] types and
    include rudimentary time zone handling through an offset in
    minutes. It is the applications responsibility to set the offset
    to correct values. The offsets are then used in date calculations.

    The implementation has not yet been thoroughly tested, but
    provides a good example of the swiftness with which you can build
    new date/time classes on top of the two basic types. If you find
    any errors or would like to see new features, mail them to
    mal@lemburg.com.

"""
__version__ = '0.1alpha'
__author__ = 'Marc-Andre Lemburg, mailto:mal@lemburg.com'

import DateTime

class _EmptyClass: pass

class Date:

    offset = 0                  # from some imaginary time zone in minutes

    def __init__(self,*args):

        self.data = apply(DateTime.DateTime,args)

    def set_timezone(self,offset):

        self.offset = offset

    def __getattr__(self,what):

        return getattr(self.data,what)

    def __sub__(self,other):

        if isinstance(other,Date):
            if self.offset != other.offset:
                # Be careful about different offsets:
                d = (self.data - self.offset * DateTime.oneMinute) \
                    - (other.data - other.offset * DateTime.oneMinute)
            else:
                # Offsets are equal: no adjustment needed
                d = self.data - other.data
            o = _EmptyClass()
            o.__class__ = Interval
            o.data = d
            return o
        elif isinstance(other,Interval):
            d = self.data - other.data
            o = _EmptyClass()
            o.__class__ = Date
            o.data = d
            o.offset = self.offset # inherit the offset
            return o
        else:
            raise TypeError,"operation not supported"

    def __add__(self,other):

        if isinstance(other,Interval):
            d = self.data + other.data
            o = _EmptyClass()
            o.__class__ = Date
            o.data = d
            o.offset = self.offset # inherit the offset
            return o
        else:
            raise TypeError,"operation not supported"

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return '<Date object for "%s" at %x>' % (str(self.data),id(self))

class Timestamp(Date):

    def __repr__(self):
        return '<Timestamp object for "%s" at %x>' % (str(self.data),id(self))

class Time:

    def __init__(self,*args):

        self.data = apply(DateTime.TimeDelta,args)

    def __getattr__(self,what):

        return getattr(self.data,what)

    def __sub__(self,other):

        if isinstance(other,Time):
            d = self.data - other.data
            o = _EmptyClass()
            o.__class__ = Interval
            o.data = d
            return o
        elif isinstance(other,Interval):
            d = self.data - other.data
            o = _EmptyClass()
            o.__class__ = Time
            o.data = d
            return o
        else:
            raise TypeError,"operation not supported"

    def __add__(self,other):

        if isinstance(other,Time):
            d = self.data + other.data
            o = _EmptyClass()
            o.__class__ = Interval
            o.data = d
            return o
        elif isinstance(other,Interval):
            d = self.data + other.data
            o = _EmptyClass()
            o.__class__ = Time
            o.data = d
            return o
        else:
            raise TypeError,"operation not supported"

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return '<Time object for "%s" at %x>' % (str(self.data),id(self))

class Interval:

    def __init__(self,*args):

        self.data = apply(DateTime.DateTimeDelta,args)

    def __getattr__(self,what):

        return getattr(self.data,what)

    def __sub__(self,other):

        if isinstance(other,Interval):
            d = self.data - other.data
            o = _EmptyClass()
            o.__class__ = Interval
            o.data = d
            return o
        else:
            raise TypeError,"operation not supported"

    def __add__(self,other):

        if isinstance(other,Interval):
            d = self.data + other.data
            o = _EmptyClass()
            o.__class__ = Interval
            o.data = d
            return o
        else:
            raise TypeError,"operation not supported"

    def __mul__(self,other):

        value = float(other)
        d = value * self.data
        o = _EmptyClass()
        o.__class__ = Interval
        o.data = d
        return o
    __rmul__ = __mul__

    def __div__(self,other):

        value = float(other)
        d = self.data / value
        o = _EmptyClass()
        o.__class__ = Interval
        o.data = d
        return o

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return '<Interval object for "%s" at %x>' % (str(self.data),id(self))

if __name__ == '__main__':
    # Some test instances to play around with
    d = Date(1998,3,2)
    e = Date(1998,1,2)
    f = Date(1998,1,2)
    f.set_timezone(60)
    t = Time(12,0,0)
    u = Time(13,0,0)



#!/usr/local/bin/python -u

import mx.DateTime            
from mx.DateTime import *
from mx.DateTime import __version__
import time,sys,traceback

print 'Testing mxDateTime version',__version__
print

if not __debug__:
    print '*** You have to run the test suite in non-optimized mode !'
    print
    sys.exit(1)

### Constructors

## DateTime()

d = DateTime(2008,1,2, 12,13,14)
assert d.year == 2008
assert d.month == 1
assert d.day == 2
assert d.hour == 12
assert d.minute == 13
assert d.second == 14
assert d.absdate == 733043
assert d.abstime == 43994.0
assert long(d.absdays*1000) == 733042509L

# Normalization
assert DateTime(2008,1,-1).day == 31
assert DateTime(2008,1,-31).day == 1
assert DateTime(2008,-1,1).month == 12
assert DateTime(2008,-12,1).month == 1
assert DateTime(-1,1,1).year == -1

# .rebuild()
assert id(d) != id(d.rebuild())
assert d.rebuild(year=2000) == DateTime(2000,1,2, 12,13,14)
assert d.rebuild(month=11) == DateTime(2008,11,2, 12,13,14)
assert d.rebuild(day=12) == DateTime(2008,1,12, 12,13,14)
assert d.rebuild(hour=9) == DateTime(2008,1,2, 9,13,14)
assert d.rebuild(minute=20) == DateTime(2008,1,2, 12,20,14)
assert d.rebuild(second=33) == DateTime(2008,1,2, 12,13,33)
assert d.rebuild(second=35.5) == DateTime(2008,1,2, 12,13,35.5)
assert d.rebuild(year=2000, hour=10) == DateTime(2000,1,2, 10,13,14)

## Test DateTimeFromAbsDateTime()

# wintertime
t = winter = DateTimeFromAbsDateTime(729368,55931.522913)
print t
print repr(t)
assert str(t) == '1997-12-10 15:32:11.52'

x = float(t)
print 'as ticks:',x
print 'time.gmtime:',time.gmtime(x),'(note the seconds)'
print 'time.localtime:',time.localtime(x)
print 'tuple:',t.tuple()

print

# summertime
t = summer = DateTimeFromAbsDateTime(729200,55931.522913)
print t
print repr(t)
assert str(t) == '1997-06-25 15:32:11.52'

x = float(t)
print 'as ticks:',x
print 'time.gmtime:',time.gmtime(x),'(note the seconds)'
print 'time.localtime:',time.localtime(x)
print 'tuple:',t.tuple()

# Leap seconds
try:
    leap_2008_0 = DateTimeFromAbsDateTime(733407, 86401.75)
except RangeError:
    pass
else:
    raise AssertionError('DateTimeFromAbsDateTime')
leap_2008_1 = DateTimeFromAbsDateTime(733407, 86400.75)
assert leap_2008_1.abstime == 86400.75
assert str(leap_2008_1) == '2008-12-31 23:59:60.75'
leap_2008_2 = DateTimeFromAbsDateTime(733407, 86400.99)
assert leap_2008_2.abstime == 86400.99
assert str(leap_2008_2) == '2008-12-31 23:59:60.99'
leap_2008_2 = DateTimeFromAbsDateTime(733407, 86400.999)
assert leap_2008_2.abstime == 86400.999
assert str(leap_2008_2) == '2008-12-31 23:59:60.99'

print

## Test Timestamp()
t = Timestamp(1900,1,1,12,23,34.5)
print t
assert t.tuple()[:6] == (1900,1,1,12,23,34)
assert str(t) == '1900-01-01 12:23:34.50'

## Test DateTimeFromCOMDate()
t = DateTimeFromCOMDate(12345.6)
print t
assert t.COMDate() == 12345.6
assert str(t) == '1933-10-18 14:24:00.00'

t = DateTimeFromCOMDate(-12345.6)
print t
assert t.COMDate() == -12345.6
assert str(t) == '1866-03-13 14:24:00.00'

## Test DateTimeFromTicks()
x = time.time()
print 'ticks:',x,'localtime:',time.localtime(x)
t = DateTimeFromTicks(x)
print t.tuple(), t.absdate, t.abstime
assert t.ticks() == x
print

## Test Date()
t = Date(2007,1,1)
#print t
assert str(t) == '2007-01-01 00:00:00.00'
assert t.absdate == 732677
assert t.date == '2007-01-01'
assert t.time == '00:00:00.00'

# Negative days count from the month's end
t = Date(2007,1,-1)
#print t
assert str(t) == '2007-01-31 00:00:00.00'
assert t.absdate == 732707
assert t.date == '2007-01-31'
assert t.time == '00:00:00.00'
t = Date(2007,2,-1)
#print t
assert t.absdate == 732735
assert t.date == '2007-02-28'
assert t.time == '00:00:00.00'

## Test now()
t1 = now()
time.sleep(0.1)
t2 = now()
if t1 == t2:
    print 'WARNING: now() does not provide sub-second accuracy'
else:
    # Report now_resolution
    if hasattr(mx.DateTime, 'now_resolution'):
        print ('now() resolution reported by the system: '
               '%f microseconds' % (now_resolution * 1e6))

    # Test now() resolution
    min_t_diff = 1.0
    current_time = now
    for i in range(1000):
        t1, t2 = current_time(), current_time()
        t_diff = abs(t2.abstime - t1.abstime)
        if t_diff < min_t_diff:
            min_t_diff = t_diff
    if min_t_diff > 0:
        print ('now() resolution measured as < %f microseconds'
               % (min_t_diff * 1e6))
    else:
        print 'now() resolution cannot be determined'


print
print 'Basic constructors ok.'

### String format

# Negative years
t = DateTime(-1,1,1,12,34,56.78)
assert str(t) == '-0001-01-01 12:34:56.78'

# Rounding
t = DateTime(2007,1,1,12,34,56.78)
assert str(t) == '2007-01-01 12:34:56.78'
t = DateTime(2011,1,1,12,34,3.42)
# Note: 3.42 is stored as 3.4199999999999999
assert str(t) == '2011-01-01 12:34:03.42', str(t)
assert repr(t).find( '2011-01-01 12:34:03.42') > 0
assert t.date == '2011-01-01'
assert t.time == '12:34:03.42'
t = DateTime(2011,1,1,12,34,3.425)
# Note: 3.425 is stored as 3.4249999999999998
assert str(t) == '2011-01-01 12:34:03.43', str(t)
assert repr(t).find( '2011-01-01 12:34:03.43') > 0
assert t.date == '2011-01-01'
assert t.time == '12:34:03.43'
t = DateTime(2011,1,1,12,34,56.1147)
# Note: 56.1147 is stored as 56.114699999999999
assert str(t) == '2011-01-01 12:34:56.11', str(t)
assert repr(t).find( '2011-01-01 12:34:56.11') > 0
assert t.date == '2011-01-01'
assert t.time == '12:34:56.11'
t = DateTime(2011,1,1,12,34,56.124449)
# Note: 56.124449 is stored as 56.124448999999998
assert str(t) == '2011-01-01 12:34:56.12', str(t)

# Rounding at the end of a minute
t = DateTimeFromAbsDateTime(733407, 86399)
assert str(t) == '2008-12-31 23:59:59.00'
t = DateTimeFromAbsDateTime(733407, 86399.9)
assert str(t) == '2008-12-31 23:59:59.90'
t = DateTimeFromAbsDateTime(733407, 86399.99)
assert str(t) == '2008-12-31 23:59:59.99'
assert repr(t).find('2008-12-31 23:59:59.99') > 0
t = DateTimeFromAbsDateTime(733407, 86399.995)
assert str(t) == '2008-12-31 23:59:59.99'
assert repr(t).find('2008-12-31 23:59:59.99') > 0
t = DateTimeFromAbsDateTime(733407, 86399.999)
assert str(t) == '2008-12-31 23:59:59.99'
assert repr(t).find('2008-12-31 23:59:59.99') > 0

print 'String format ok.'

### Python Number protocol:
t1 = Date(1997,12,31)
t2 = Date(1997,12,30)
assert int(t1) == int(t1.ticks())
d1 = t1 - t2
assert int(d1) == 86400
d2 = t2 - t1
assert int(d2) == -86400
try:
    t1 + t2
except TypeError:
    pass
else:
    raise TypeError('DateTime+DateTime should not work')
try:
    t2 * t1
except TypeError:
    pass
else:
    raise TypeError('DateTime*DateTime should not work')
try:
    t2 / t1
except TypeError:
    pass
else:
    raise TypeError('DateTime/DateTime should not work')

print 'Number protocol ok.'

### Deltas

# Constructors
assert DateTimeDelta(2,1,-1,-3).tuple() == (2, 0, 58, 57.0)
assert DateTimeDelta(2,0,-1,-3).tuple() == (1, 23, 58, 57.0)

# .rebuild()
t = DateTimeDelta(2, 12, 13, 14)
assert id(t) != id(t.rebuild())
assert t.rebuild(day=2000) == DateTimeDelta(2000, 12,13,14)
assert t.rebuild(hour=9) == DateTimeDelta(2, 9, 13, 14)
assert t.rebuild(minute=20) == DateTimeDelta(2, 12, 20, 14)
assert t.rebuild(second=33) == DateTimeDelta(2, 12, 13, 33)
assert t.rebuild(second=35.5) == DateTimeDelta(2, 12, 13, 35.5)
assert t.rebuild(hour=20, second=10) == DateTimeDelta(2, 20, 13, 10)

# DateTime calculations
assert DateTime(1998,1,9) + DateTimeDelta(-1) == DateTime(1998,1,8)
assert DateTime(1998,1,9) - DateTimeDelta(1) == DateTime(1998,1,8)
assert DateTime(1998,1,9) - DateTimeDelta(1.5) == DateTime(1998,1,7,12,0,0)
assert DateTime(1998,1,9) - DateTime(1969,4,6) == DateTimeDelta(10505)
assert DateTime(2000,12,31) + DateTimeDelta(1) == DateTime(2001, 1, 1)
assert DateTime(2000,12,31) + DateTimeDelta(2) == DateTime(2001, 1, 2)

# Integer calculations
assert DateTime(2000,12,31) + 1 == DateTime(2001, 1, 1)
assert DateTime(2000,12,31) + 2 == DateTime(2001, 1, 2)

# DateTimeDelta calculations
assert DateTimeDelta(10000) / (DateTimeDelta(10000) * 0.5) == 2.0

# Attributes
a = Date(2000,11,9)
b = Date(2000,11,10)
d1 = b - a
d2 = a - b
assert d1.seconds == 86400.0
assert d1.day == 1
assert d1.hour == 0
assert d1.second == 0
assert d2.seconds == -86400.0

# Number protocol
assert int(d1) == 86400
assert int(d2) == -86400

assert int(d1 * 3) == 86400 * 3
assert int(3 * d1) == 86400 * 3

assert int(d1 + d2) == 0
assert int(d2 + d1) == 0
assert int(d1 - d2) == 86400 * 2, d1 - d2
assert int(d2 - d1) == -86400 * 2, d2 - d1

try:
    d2 * d1
except TypeError:
    pass
else:
    raise TypeError('DateTimeDelta*DateTimeDelta should not work')

print 'Deltas ok.'
print

### UTC vs. local time
t = DateTime(2000,11,17)
print 'UTC -> local time -> UTC'
for i in range(24*365):
    x = t + 0.041666*i

    try:
        try:
            assert cmp(x.gmtime().localtime(), x, 0.001) == 0
        except AssertionError:
            print '    Problem (local time) -- DST switching time ?:'
            print '    %s -> %s -> %s' % \
                  (x, x.gmtime(), x.gmtime().localtime())

        try:
            assert cmp(x.localtime().gmtime(), x, 0.001) == 0
        except AssertionError:
            print '    Problem (gmt time) -- DST switching time ?:'
            print '    %s -> %s -> %s' % \
                  (x, x.localtime(), x.localtime().gmtime())

    except Error, why:
        print '*** Problem: %s' % x
        
print 'ok.'

### Rounding problems with COM dates
t = DateTime(1998, 2, 24, 12, 40, 11)
c = DateTimeFromCOMDate(t.COMDate())
print 'DateTime->COM Date->DateTime rounding error:',(t == c) * 'no' or 'yes'
print 'COM Dates compare:',(t.COMDate()==c.COMDate()) * 'equal' or 'not equal'
print 'diff =',t-c,'in seconds:',t.second - c.second
print 'using cmp(,,0.5):',(cmp(t,c,0.5) == 0) * 'equal' or 'not equal'
print

### Julian calendar

t = JulianDateTime(1752, 9, 2)
assert t.tuple()[:3] == (1752, 9, 2)
assert t.date == '1752-09-02'
assert t.calendar == 'Julian'
g = t.Gregorian()
assert g.tuple()[:3] == (1752, 9, 13)
assert g.date == '1752-09-13'
assert g.calendar == 'Gregorian'

### Constructor tests

def testjdn(year,month,day):

    # Reference taken from the Calendar FAQ (see docs for reference)

    a = (14-month)/12
    y = year+4800-a
    m = month + 12*a - 3
    
    # Gregorian
    gJDN = day + (306*m+5)/10 + y*365 + y/4 - y/100 + y/400 - 32045
    # Julian
    jJDN = day + (306*m+5)/10 + y*365 + y/4 - 32083
    
    return gJDN,jJDN

print 'Running constructor test... (this can take up to a few minutes)'
try:
    if __version__ >= '1.3.0':
        # New versions:
        for suite in (range(-100,100),range(1900,2101)):
            p = None
            for year in suite:
                print year,; sys.stdout.flush()
                for month in range(1,13):
                    for day in range(1,32):
                        try:
                            t = DateTime(year,month,day,12)
                            jt = JulianDateTime(year,month,day,12)
                            ref_greg,ref_jul = testjdn(year,month,day)
                        except RangeError:
                            continue
                        if p:
                            try:
                                assert p + 1 == t
                                assert (p+1).tuple() == t.tuple()
                                assert t.tuple()[:3] == (year,month,day)
                                assert DateTimeFromAbsDateTime(t.absdate).tuple()[:3] \
                                       == (year,month,day)
                                assert t.jdn == ref_greg
                                assert jt.jdn == ref_jul
                            except AssertionError:
                                print '*** Test failed for:'
                                print '  p = %s, p+1 = %s and' % (p,p+1)
                                print '  t = %s, jdn=%i; jdn reference = %i' % \
                                      (t,t.jdn,ref_greg)
                                print '  jt = %s, jdn=%i; jdn reference = %i' % \
                                      (jt,jt.jdn,ref_jul)
                                print
                                traceback.print_exc()
                                print
                                sys.exit(1)
                        p = t
            print
    else:
        # Old versions and experimental Python version:
        for suite in (range(1900,2101),):
            p = None
            for year in suite:
                print year,; sys.stdout.flush()
                for month in range(1,13):
                    for day in range(1,32):
                        try:
                            t = DateTime(year,month,day,12)
                            ref_greg,ref_jul = testjdn(year,month,day)
                        except RangeError:
                            continue
                        if p:
                            try:
                                assert p + 1 == t
                                assert (p+1).tuple() == t.tuple()
                                assert t.tuple()[:3] == (year,month,day)
                                assert DateTimeFromAbsDateTime(t.absdate).tuple()[:3] \
                                       == (year,month,day)
                                assert t.jdn == ref_greg
                            except AssertionError:
                                print '*** Test failed for:'
                                print '  p = %s, p+1 = %s and' % (p,p+1)
                                print '  t = %s, jdn=%i; jdn reference = %i' % \
                                      (t,t.jdn,ref_greg)
                                print
                                traceback.print_exc()
                                print
                                sys.exit(1)
                        p = t
            print
except KeyboardInterrupt:
    print
    print 'Interrupted.'
else:
    print 'Date construction works.'
    print

# Non zero testing
if t:
    print 'Non zero testing works.'

# ticks and dst
print 'Converting ticks and DST handling.'
try:
    summer.ticks(0,0)
    winter.ticks(0,0)
    summer.ticks(0,1)
    winter.ticks(0,1)
except SystemError:
    print
    print '-'*72
    print 'WARNING:'
    print
    print 'The mktime() C API on your platform does not support'
    print 'setting the DST flag to anything other than -1. This'
    print 'will cause the datetime.ticks() method to raise an'
    print 'error in case you pass a DST flag other than -1.'
    print '-'*72

# Check for bug #707
epoch = DateTimeFromTicks(0)
one_second_before_epoch = epoch - oneSecond
two_seconds_before_epoch = epoch - 2*oneSecond
try:
    assert two_seconds_before_epoch.ticks() == -2
except Error:
    print 'Negative ticks are not support by the C lib mktime().'
else:
    assert one_second_before_epoch.ticks() == -1
    print 'Negative ticks are supported by the C lib mktime().'

# Sanity checks
date1 = Date(2000,11,1) + \
        RelativeDateTime(day=1,month=1,hour=0,minute=0,second=0)
date2 = Date(2000,11,1) + \
        RelativeDateTime(day=1,month=6,hour=0,minute=0,second=0)

print '%.15f == %.15f ?' % (date1.ticks(), time.mktime(date1.tuple()))
print '%.15f == %.15f ?' % (date2.ticks(), time.mktime(date2.tuple()))

try:
    assert date1.ticks() == time.mktime(date1.tuple())
    assert date2.ticks() == time.mktime(date2.tuple())
except AssertionError:
    print
    print '-'*72
    print 'WARNING:'
    print
    print 'Conversion from mx.DateTime instances to ticks is not relyable'
    print 'on your platform. Please run "python testticks.py" and send'
    print 'the output to the author at mal@lemburg.com.'
    print
    print '-'*72

print '%s == %s ?' % \
      (time.localtime(time.mktime(date1.tuple()))[:6], 
      date1.tuple()[:6])
print '%s == %s ?' % \
      (time.localtime(time.mktime(date2.tuple()))[:6], 
      date2.tuple()[:6])

try:
    assert time.localtime(time.mktime(date1.tuple()))[:6] == \
           date1.tuple()[:6]
    assert time.localtime(time.mktime(date2.tuple()))[:6] == \
           date2.tuple()[:6]
except AssertionError:
    print
    print '-'*72
    print 'WARNING:'
    print
    print 'time.localtime() and time.mktime() are not inverse functions'
    print 'on your platform. This may lead to strange results in some'
    print 'applications. Please run "python testticks.py" and send'
    print 'the output to the author at mal@lemburg.com.'
    print
    print '-'*72

print

# RelativeDateTime
print 'RelativeDateTime...',
assert hash(RelativeDateTime(year=2002)) == 135051564
assert hash(RelativeDateTime(year=2001)) == 135051567
assert hash(RelativeDateTime(year=2004)) == 135051562
assert RelativeDateTime(year=2004) == RelativeDateTime(year=2004)
assert RelativeDateTime(year=2004) != RelativeDateTime(year=2002)
assert str(RelativeDateTime(minutes=75)) == 'YYYY-MM-DD HH:(+75):SS'
assert str(RelativeDateTime(minutes=-75)) == 'YYYY-MM-DD HH:(-75):SS'
assert str(RelativeDateTime(hours=0.5)) == 'YYYY-MM-DD (+00):(+30):SS'
assert str(RelativeDateTime(hours=-0.5)) == 'YYYY-MM-DD (+00):(-30):SS'
print 'done.'

# DateTimeFrom()
print 'DateTimeFrom()...',
assert (DateTimeFrom(hour=12, minute=3, defaultdate=today()-1) ==
        today() - 1 + TimeDelta(12, 3))
assert (DateTimeFrom(month=12, day=1, defaultdate=DateTime(2009, 1, 1, 12)) ==
        DateTime(2009, 12, 1, 12))
assert (DateTimeFrom("12:03", defaultdate=today()-1) ==
        today() - 1 + TimeDelta(12, 3))
#assert (DateTimeFrom("Dec 1", defaultdate=DateTime(2009, 1, 1, 12)) ==
#        DateTime(2009, 12, 1, 12))
assert (DateTimeFrom("Dec 1", defaultdate=DateTime(2009, 1, 13)) ==
        DateTime(2009, 12, 1))
print 'done.'

# gmtime()
print
print 'Testing gmtime()...',
t = 2**31 - 100
for t, refdate in ((2**31, DateTime(2038, 1, 19, 3, 14, 8)),
                   (2**32, DateTime(2106, 2, 7, 6, 28, 16)),
                   (2**33, DateTime(2242, 3, 16, 12, 56, 32)),
                   (2**34, DateTime(2514, 5, 30, 1, 53, 4)),
                   ):
    for i in range(-100, 100):
        try:
            x = gmtime(t + i)
            assert x == refdate + i*oneSecond, (t, i, refdate)
        except ValueError, reason:
            print 'Failed for %s: %s + %i seconds' % (t + i, refdate, i)
            break
print 'done.'

# timegm() emulation
print
print 'Testing .gmticks()... (this can take up to a few minutes)'
t = start = 920710000
stop = 2140240000
oops = 0
try:
    while 1:
        if t % 10000 < 20:
            print t,
        d = apply(DateTime,time.gmtime(t)[:6])
        try:
            x = d.gmticks()
        except Error:
            break
        if x != t:
            print ' Ooops:',d,'t =',t,'diff =',x-t
            oops = oops + 1
        try:
            t = t + 10011
        except OverflowError:
            break
        else:
            if t > stop:
                break
except KeyboardInterrupt:
    print
    print 'Interrupted.'
else:
    print
print ' Tested ticks range %i to %i.' % (start,t)
assert oops == 0
print '...Works.'
print

# Test PyDateTime integration
print 'Running Python datetime module integration tests.'
from mx.DateTime.mxDateTime import testpydatetime
print

# Run slot operations tests
print 'Running slot operation tests.'
from mx.DateTime.mxDateTime import testslotops
print

# Try importing a subpackage
print 'Importing subpackage Feasts.'
print
from mx.DateTime import Feasts
try:
    Feasts._test()
except SystemError:
    print '-'*72
    print 'WARNING:'
    print
    print 'Subpackges ISO and ARPA will not work on your platform because'
    print 'mxDateTime found no working API to query the timezone'
    print 'for a given date/time. Please run "python testticks.py" and send'
    print 'the output to the author at mal@lemburg.com.'
    print
    print '-'*72
print

# Run parser tests
print 'Importing subpackage Parser.'
print
from mx.DateTime import Parser
Parser._test()

print
print 'Works.'



""" Namespace objects.

    Copyright (c) 1998-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
    Copyright (c) 2000-2013, eGenix.com Software GmbH; mailto:info@egenix.com
    See the documentation for further information on copyrights,
    or contact the author. All Rights Reserved.
"""
import types
import mx.Tools.NewBuiltins
from mx.Tools import freeze, attributes

class Namespace:

    """ Base class for Namespaces.

        Instances of this namespace can store arbitrary attributes,
        which are stored in the instances dictionary. They are then
        available via attribute lookup, e.g. ns.attr, and index
        lookup, e.g. ns['attr'].

        The Namespaces support acquisition. To link the object to
        another object, set self.baseobj to the other
        object. Attributes not found in the Namespace instance are
        then looked up in the other object as well.

        NOTE:

        All methods and internal attributes must start with an
        underscore. Namespace attributes must *not* start with an
        underscore to prevent naming collisions with internal
        variables and methods.  Names with leading underscore can
        *not* be acquired.

    """

    def __init__(self, **kws):

        """ Initializes the namespace with the attributes given
            as keywords.
        """
        if kws:
            self.__dict__.update(kws)
    
    def _clear(self,save=(),recursive=0,

               getattr=getattr):

        """ Clears the namespace.

            The attributes named in save are not cleared. Per default
            all attributes are cleared. You may want to save the
            baseobj attribute in case you use it.

            Class variables are not cleared this way, but overriding
            instance variables do get deleted.

            If recursive is true (default is false), the namespace is
            cleared recursively, meaning that all Namespace instances
            contained in it are also cleared.

        """
        d = {}
        for attr in save:
            d[attr] = getattr(self,attr)
        # Clear the old instance dictionary (recursively) so that it
        # can be garbage collected
        if recursive:
            for obj in self.__dict__.items():
                if isinstance(obj, Namespace):
                    obj._clear()
        obj = None
        self.__dict__.clear()
        # Assing the new one
        self.__dict__ = d

    def _update(self, other_namespace,

                DictType=types.DictType,type=type):

        """ Updates the namespace dictionary with the new entries
            taken from the object other_namespace.

            other_namespace can either be a dictionary or another
            Namespace object.

        """
        if type(other_namespace) is DictType:
            self.__dict__.update(other_namespace)
        else:
            self.__dict__.update(other_namespace.__dict__)

    def _load(self, other_namespace, save=(),
              ignore_types=(types.BuiltinMethodType,
                            types.ClassType,
                            types.MethodType,
                            types.FunctionType),

              DictType=types.DictType,type=type):

        """ Updates the namespace with the updated entries taken from
            the object other_namespace provided that the entries already
            exist in the namespace dictionary

            other_namespace can either be a dictionary or another
            Namespace object.

            Attributes listed in save are not overwritten. New values
            having one of the types listed in ignore_types are also
            ignored.

        """
        if type(other_namespace) is DictType:
            items = other_namespace.items()
        else:
            items = other_namespace._items()
        d = self.__dict__
        exists = d.has_key
        for k, v in items:
            if exists(k):
                if k in save or \
                   type(v) in ignore_types:
                    continue
                d[k] = v
        exists = None
        d = None

    def _load_file(self, filename):

        """ Load the namespace from the file filename.
        """
        execfile(filename, self.__dict__)

    def _copy_to(self, other_namespace,

                 DictType=types.DictType,type=type):

        """ Copy the namespace contents into the dictionary or Namespace
            instance other_namespace.
            
        """
        if type(other_namespace) is DictType:
            other_namespace.update(self.__dict__)
        else:
            other_namespace._update(self.__dict__)

    def _format(self, template, recursive=0):

        """ Format the text template using the namespace contents.

            template must be a text using Python string placeholders,
            e.g. "%(attrname)s" or "%(attrname)r". Formatting is done
            using all available attributes of the namespace (including
            class attributes).

            If recursive is true, the formatting is applied recursively
            (as long as the string '%(' occurs in the template), i.e.
            the Namespace attributes may also contain Python string
            placeholders.
            
        """
        if not recursive:
            return template % attributes(self)
        dict = attributes(self)
        while template.find('%(') >= 0:
            template = template % dict
        return template

    def _attributes(self):

        """ Return all available attributes as dictionary.

            The dictionary includes all class and instance attributes.
            
        """
        return attributes(self)

    def _dictionary(self):

        """ Returns the current namespace contents as dictionary.

            The dictionaries values should be considered read-only in
            case they are mutable.

        """
        return self.__dict__.copy()

    def _copy(self):

        """ Returns a shallow copy of the current namespace.

            See the standard Python module "copy" for details on the
            copying mechanism.

        """
        import copy
        return copy.copy(self)

    def _deepcopy(self):

        """ Returns a deep copy of the current namespace.

            See the standard Python module "copy" for details on the
            copying mechanism.

        """
        import copy
        return copy.deepcopy(self)

    def _clone(self):

        """ Returns a shallow copy of the instance.

            This is different from the ._copy() method in that
            ._update() is used to copy the contents of the instance to
            the new instance.
        
        """
        obj = self.__class__()
        obj._update(self.__dict__)
        return obj

    ### Index interface

    def __getitem__(self,name,

                    getattr=getattr):

        """ Index interface to the namespace.

            Raises IndexErrors in case an attribute is not found.

        """
        try:
            return getattr(self, name)
        except AttributeError,why:
            raise IndexError,why
        
    def __setitem__(self,name,value,

                    setattr=setattr):

        """ Index interface to the namespace.
        """
        return setattr(self,name,value)

    def __len__(self):

        """ Return the number of instance attributes stored in the
            namespace.
        """
        return len(self.__dict__)

    def __nonzero__(self):

        """ Returns 1/0 depending on whether any instance attributes
            are defined in the namespace.

            An empty Namespace is false, if it contains at least one
            entry it is true.
            
        """
        return (len(self.__dict__) > 0)

    ### Dictionary like interface to the instance dictionary

    def _items(self):

        """ Returns a list of tuples (attrname, attrvalue) for
            all (instance) attributes stored in the object.
        """
        return self.__dict__.items()

    def _keys(self):

        """ Returns a list of attribute names for
            all (instance) attributes stored in the object.
        """
        return self.__dict__.keys()

    def _values(self):

        """ Returns a list of attribute values for
            all (instance) attributes stored in the object.
        """
        return self.__dict__.values()

    def _has_key(self, key):

        """ Returns 1/0 depending on whether the object has
            an instance attribute key or not.

            Note that class attributes cannot be queried this way.
            
        """
        return self.__dict__.has_key(key)

    # Alias
    _hasattr = _has_key

    def _get(self,key,default=None):

        """ Returns the value of attribute key, or default if no
            such attribute is defined.
        """
        return self.__dict__.get(key,default)

    # Alias
    _getattr = _get

    ### Additional methods:

    def _extract(self,names,defaults=(),

                 extract=extract):

        """ Extract a list of named objects from the namespace.
        
            defaults may be given as sequence of the same size as
            names to provide default values in case names are not
            found in the namespace.

            Only the instance dictionary is scanned, the class
            variables are not available through this mechanism.

        """
        return extract(self.__dict__,names,defaults)

    def _prefixgroup(self,prefix,nonzero=0,wholename=0):

        """ Find all attributes having prefix as common prefix and return
            a dictionary mapping the attribute suffixes (the prefix
            is removed) to their values.
            
            If nonzero is set, then only fields with non-zero value
            are included in the dictionary

            If wholename is true, the prefix is not removed from the
            attribute names.

        """
        plen = len(prefix)
        d = {}
        if not nonzero and not wholename:
            # Shortcut for the default settings
            for k,v in self.__dict__.items():
                if k[:plen] == prefix:
                    d[k[plen:]] = v
        else:
            for k,v in self.__dict__.items():
                if k[:plen] == prefix:
                    if nonzero and not v:
                        continue
                    if wholename:
                        d[k] = v
                    else:
                        d[k[plen:]] = v
        return d

    ### For debugging purposes:

    def __repr__(self,

                 id=id):

        keys = self.__dict__.keys()
        keys.sort()
        return '<%s.%s (%s) at 0x%x>' % (
            self.__class__.__module__,
            self.__class__.__name__,
            ','.join(keys),
            id(self))

    def __str__(self,

                id=id):

        items = self.__dict__.items()
        items.sort()
        l = ['%s.%s at 0x%x:' % (self.__class__.__module__,
                                 self.__class__.__name__,
                                 id(self))]
        for k,v in items:
            try:
                s = repr(v)
            except (ValueError, TypeError, AttributeError):
                s = '*** repr(value) failed ***'
            else:
                if len(s) > 200:
                    s = s[:200] + '... (truncated, full length = %i)' % len(s)
            l.append('    %-10s %s' % (k + ':', s))
        return '\n'.join(l)

###

class DefaultMixin:

    """ Mixin that provides default attribute values.

        The namespace returns _default for all undefined attributes
        (except the ones that start with an underscore, see
        __getattr__). The _default is defined as '' for this class.

        The acquisition feature is *not* available for instances of this
        class.

    """
    _default = ''

    def __getattr__(self,name):

        """ Return self._default for all undefined attributes that don't
            start with an underscore.

            For attributes that start with an underscore, raise an
            AttributeError. This is especially important if the
            objects are to be pickled, since the pickle mechanism
            queries a few methods on the objects, which it then calls.
            
        """
        if name[0] == '_':
            raise AttributeError,name
        return self._default

###

class NamespaceWithDefault(DefaultMixin,Namespace):

    """ Namespace class with an infinite number of attributes.

        The namespace returns _default for all undefined attributes
        (except the ones that start with an underscore, see
        __getattr__). The _default is defined as '' for this class.

        The acquisition feature is *not* available for instances of this
        class.

    """

freeze(NamespaceWithDefault)

###

class AcquisitionMixin:

    """ Mixin that acquires unknown attributes from a .baseobj
        (if set)

    """
    baseobj = None              # Base object ("containing" this object)

    # Implicit acquisition
    __getattr__ = acquire       # Use the (new) builtin acquire() to implement
                                # the acquisition mechanism

###

class NamespaceWithAcquisition(AcquisitionMixin, Namespace):

    """ Namespace class with acquisition mechanism.

        The namespace fetches values for all undefined attributes
        (except the ones that start with an underscore, see
        __getattr__) from the .baseobj, if defined.

    """
    
freeze(NamespaceWithAcquisition)

###

class _NotFound:
    pass
_NotFound = _NotFound()

class FormatNamespaceMixin:

    """ This mixin makes a Namespace class conform to the Python
        format string mechanism.

        The following format specifiers are supported:

        %(attribute)s		- gives the string representation
        			  of .attribute
        %(<eval-string>)s	- evaluates the <eval-string> in the
        			  Namespace dictionary and returns
                                  the string representation

        All standard Python string formatting output types are
        supported, provided that the attribute or evaluation result
        can be converted to these types.
                                  
    """

    def __getitem__(self, item,

                    _NotFound=_NotFound):

        obj = getattr(self, item, _NotFound)
        if obj is not _NotFound:
            return obj
        return eval(item, self.__dict__)

###

class FormatNamespace(FormatNamespaceMixin,
                      Namespace):

    """ Namespace class which supports the Python format string
        lookup mechanism.

        These Namespaces can be used in 'format string' % namespace
        formatting.

        The following format specifiers are supported:

        %(attribute)s		- gives the string representation
        			  of .attribute
        %(<eval-string>)s	- evaluates the <eval-string> in the
        			  Namespace dictionary and returns
                                  the string representation
                                  
        All standard Python string formatting output types are
        supported, provided that the attribute or evaluation result
        can be converted to these types.
                                  
    """
    pass

freeze(FormatNamespace)



#!/usr/local/bin/python

""" Standard Logging Interface

    Creates an object log, that can be used for global logging
    activities. The properties can later be changed by calling
    log.setup().
    
    It is safe to do a 'from mx.Log import *'.

    Copyright (c) 1999-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
    Copyright (c) 2000-2013, eGenix.com Software GmbH; mailto:info@egenix.com
    See the documentation for further information on copyrights,
    or contact the author. All Rights Reserved.

"""
# Be careful with these symbols since we want 'from mx.Log import *'
# to not clutter up the users namespace.
import sys, types, traceback, os
from mx import DateTime, Tools
(_sys,_DateTime,_types,_traceback,_os) = \
    (sys,DateTime,types,traceback,os)
del sys,DateTime,types,traceback,os

# Load new builtins
import  mx.Tools.NewBuiltins
del mx

# Infos & Errors (lower values indicate higher importance)
SYSTEM_ALWAYS_LOG = 0
SYSTEM_FATAL = 1
SYSTEM_PANIC = 2
SYSTEM_ERROR = 10
SYSTEM_CANCEL = 11
SYSTEM_IMPORTANT = 100
SYSTEM_WARNING = 500
SYSTEM_MESSAGE = 600
SYSTEM_UNIMPORTANT = 3000
SYSTEM_INFO = 1000
SYSTEM_DEBUG = 2000
# The highest used value:
SYSTEM_ANY = 9999
# Some log level limits
SYSTEM_LOG_NOTHING = 0
SYSTEM_LOG_EVERYTHING = SYSTEM_ANY

### Helpers

_NAME_LEN_LIMIT = 80
_VALUE_LEN_LIMIT = 500
_TRACEBACK_LIMIT = 100
_DICT_LEN_LIMIT = 50
_NON_RECURSIVE = (_types.ModuleType,
                  _types.FunctionType,
                  _types.MethodType,
                  _types.BuiltinFunctionType,
                  _types.BuiltinMethodType,
                  _types.ClassType,
                  __builtins__)
_STRING_TYPES = (_types.StringType,
                 _types.UnicodeType)

def is_string(obj):

    """ Return 1/0 depending on whether obj is an instance of
        a string type.

    """
    return isinstance(obj, _STRING_TYPES)

# For Python < 2.2, isinstance() does not accept tuples of types, so
# override is_string() with a version that does not need this.
try:
    isinstance('', _STRING_TYPES)
except TypeError:
    def is_string(obj):
        for typeobj in _STRING_TYPES:
            if isinstance(obj, typeobj):
                return 1
        return 0

def print_tb_locals(tb=None, *args, **kws):

    """ Print a listing of the traceback's locals() to file.
    """
    # Walk down the traceback
    if tb is None:
        tb = _sys.exc_info()[2]
    while tb.tb_next != None:
        tb = tb.tb_next

    # Show the local variables
    apply(print_frame_locals, (tb.tb_frame,) + args, kws)

def print_tb_globals(tb=None, *args, **kws):

    """ Print a listing of the traceback's globals() to file.
    """
    # Walk down the traceback
    if tb is None:
        tb = _sys.exc_info()[2]
    while tb.tb_next != None:
        tb = tb.tb_next

    # Show the global variables
    apply(print_frame_globals, (tb.tb_frame,) + args, kws)

def print_tb_vars(tb=None, *args, **kws):

    """ Print a listing of the traceback's locals() and globals() to file.

        If locals() and globals() are the same object, only the
        globals() are printed.

    """
    # Walk down the traceback
    if tb is None:
        tb = _sys.exc_info()[2]
    while tb.tb_next != None:
        tb = tb.tb_next
    if tb.tb_frame.f_locals is not tb.tb_frame.f_globals:
        # Show the local variables
        apply(print_frame_locals, (tb.tb_frame,) + args, kws)

    # Show the global variables
    apply(print_frame_globals, (tb.tb_frame,) + args, kws)

def filter_private_attributes(key):

    """ Filters all attributes which start with an underscore.

        For use as filter argument to one of the print_xxx() APIs.
    
    """
    if is_string(key) and key[:1] == '_':
        return 0
    return 1

def print_frame_locals(frame,file=_sys.stdout,
                       indent='#  ',
                       title='### Dump of local variables:\n',
                       levels=2,
                       nonrecursive=_NON_RECURSIVE,
                       filter=None):

    """ Print a listing of the locals() in frame to file.
    """
    if title:
        file.write(title)
    print_dict(frame.f_locals,file,indent,levels,
               nonrecursive=nonrecursive,filter=filter)

def print_frame_globals(frame,file=_sys.stdout,
                        indent='#  ',
                        title='### Dump of global variables:\n',
                        levels=2,
                        nonrecursive=_NON_RECURSIVE,
                        filter=None):

    """ Print a listing of the globals() in frame to file.
    """
    if title:
        file.write(title)
    print_dict(frame.f_globals,file,indent,levels,
               nonrecursive=nonrecursive,filter=filter)

def print_dict(dict,file,indent='',levels=1,reprkeys=0,
               nonrecursive=(),filter=None):

    """ Print dictionary to file.

        reprkeys=1 will print the items using repr() for keys as well.
        indent is prepended to all lines. levels indicates the number
        of recursion levels to be printed.

        filter may be given as callable taking the dictionary key as
        input. It should then return 1 for keys which should be
        displayed and 0 for ones which should be skipped.

    """
    l = dict.items()
    l.sort()
    if len(l) > _DICT_LEN_LIMIT:
        l = l[:_DICT_LEN_LIMIT]
        l.append(('...','...truncated...'))
    for k,v in l:
        if filter is not None:
            if not filter(k):
                continue
        if reprkeys:
            try:
                n = repr(k)
            except:
                n = '*repr()-error*'
        else:
            n = str(k)
        try:
            r = repr(v)
        except:
            r = '*repr()-error*'

        # Truncate
        if len(n) > _NAME_LEN_LIMIT:
            n = n[:_NAME_LEN_LIMIT] + '...'
        if len(r) > _VALUE_LEN_LIMIT:
            r = r[:_VALUE_LEN_LIMIT] + '...'

        # Write item
        if reprkeys:
            file.write('%s %-15s = %s\n' % (indent,n,r))
        else:
            file.write('%s.%-15s = %s\n' % (indent,n,r))
        if levels > 1:
            print_recursive(v,file,indent + '  ',levels-1,
                            nonrecursive=nonrecursive, filter=filter)

def print_sequence(obj,file=_sys.stdout,indent='',levels=2,
                   nonrecursive=()):

    l = []
    unfold = 0
    try:
        length = len(obj)
    except (AttributeError, ValueError, TypeError):
        return
    
    for i in Tools.trange(min(length,_VALUE_LEN_LIMIT)):
        try:
            value = obj[i]
        except:
            break
        try:
            r = repr(value)
        except:
            r = '*repr()-error*'

        # Truncate
        if len(r) > _VALUE_LEN_LIMIT:
            r = r[:_VALUE_LEN_LIMIT] + '...'

        # Write value
        l.append((value,r))

        # Only unfold sequences that have non-string items or string items
        # with more than on character
        if not is_string(value) or len(value) > 1:
            unfold = 1
            
    if len(obj) > _VALUE_LEN_LIMIT:
        l.append(('...','...truncated...'))

    # Unfold value object
    if unfold:
        for i,(value,rvalue) in irange(l):
            file.write('%s%-15s = %s\n' % (indent, '[%i]' % i, rvalue))
            if levels > 1:
                print_recursive(value,file,indent + '  ',levels-1,
                                nonrecursive=nonrecursive)

def print_recursive(obj,file=_sys.stdout,indent='',levels=1,
                    nonrecursive=(),filter=None):

    # Filter out nonrecursive types and objects
    try:
        if type(obj) in nonrecursive or \
           obj in nonrecursive:
            return
    except:
        # Error during compares result in the object not being
        # printed
        return
    
    # Print the object depending on its interface
    if hasattr(obj,'__dict__') and \
       obj.__dict__ is not None:
        print_dict(obj.__dict__,file,indent,levels,
                   nonrecursive=nonrecursive, filter=filter)

    elif hasattr(obj,'items'):
        print_dict(obj,file,indent,levels,1,
                   nonrecursive=nonrecursive, filter=filter)

    elif Tools.issequence(obj) and not is_string(obj):
        print_sequence(obj,file,indent,levels,
                       nonrecursive=nonrecursive)

    elif hasattr(obj,'__members__'):
        d = {}
        for attr in obj.__members__:
            d[attr] = getattr(obj,attr)
        print_dict(d, file, indent, levels,
                   nonrecursive=nonrecursive, filter=filter)

def print_obj(obj,file=_sys.stdout,indent='',levels=2,
              nonrecursive=(),filter=None):

    try:
        r = repr(obj)
    except:
        r = '*repr()-error*'

    # Truncate
    if len(r) > _VALUE_LEN_LIMIT:
        r = r[:_VALUE_LEN_LIMIT] + '...'

    # Write object representation
    file.write('%s%s\n' % (indent,r))
    print_recursive(obj, file, indent + '  ', levels-1, filter=filter)

def print_stack_with_locals(file=_sys.stdout,levels=100,offset=0):

    print_stack(file, levels=levels, offset=offset+1, locals=1)

def print_stack(file=_sys.stdout,levels=100,offset=0,locals=0):

    # Prepare frames
    try:
        raise ValueError
    except ValueError:
        # Go back offset+1 frames...
        f = _sys.exc_info()[2].tb_frame
        for i in range(offset + 1):
            if f.f_back is not None:
                f = f.f_back

    # Extract frames
    frames = []
    while f:
        frames.append(f)
        f = f.f_back
    frames.reverse()

    # Prepare stack
    stack = _traceback.extract_stack()

    # Make output
    file.write('Stack:\n')
    for (frame,(filename, lineno, name, line)) in \
            Tools.tuples(frames,stack)[-levels:]:
        file.write(' File "%s", line %d, in %s\n' % (filename,lineno,name))
        if line:
            file.write('  %s\n' % line.strip())
        if locals:
            print_frame_locals(frame,file,indent='   |',title='')

def print_loaded_modules(file=_sys.stdout,

                         sys=_sys):

    """ Print a listing of currently loaded modules to file.
    """
    l = sys.modules.items()
    l.sort()
    file.write('Loaded modules and packages:\n')
    for k,v in l:
        p = k.split('.')
        for i in range(len(p)-1):
            p[i] = '   '
        n = ''.join(p)
        if v is not None:
            if hasattr(v,'__path__'):
                # package
                file.write(' %s[%s]\n' % (''.join(p[:-1]),p[-1]))
            else:
                # module
                file.write(' %s\n' % (n))

def format_traceback(title='### Dump of local variables:\n',
                     filter=None):

    """ Formats a traceback of the current exception and
        returns it as string.

        Includes a listing of the locals() causing that were active
        during the exception. title is used to separate the traceback
        from the listing of variables. filter can be given to filter
        out certain variables from the listing. It takes one argument:
        the variable name and has to return 1/0.

    """
    import cStringIO
    f = cStringIO.StringIO()
    _traceback.print_exc(_TRACEBACK_LIMIT,f)
    print_tb_locals(_sys.exc_info()[2], f, title=title, filter=filter)
    return f.getvalue()

### Log class

class Log:

    """ Log class
    
    """

    # Log ignore level. Log messages with an errno >= this ignore
    # level are not logged
    ignore_level = SYSTEM_LOG_EVERYTHING

    # Log file to use. This can be a pathname or 'stderr'/'stdout' to
    # log to the special file descriptors stderr/stdout.
    log_file = 'stderr'

    # Log ID to use in the log output. Default is set in .__init__()
    log_id = None

    # Open file-like object. This is opened on demand and can be closed
    # via .close()
    open_log_file = None

    # Line buffer used by .write()
    line_buffer = ''
    
    ### Attributes needed for file interface

    # We can always accept data via .write(). Note that this variable
    # is not changed by calling .close().
    closed = 0

    ### Convenience attributes

    # Log levels as attributes for convenience; you can then write e.g.
    # log(log.INFO, ...)
    ALWAYS_LOG = SYSTEM_ALWAYS_LOG
    FATAL = SYSTEM_FATAL
    PANIC = SYSTEM_PANIC
    ERROR = SYSTEM_ERROR
    CANCEL = SYSTEM_CANCEL
    IMPORTANT = SYSTEM_IMPORTANT
    WARNING = SYSTEM_WARNING
    MESSAGE = SYSTEM_MESSAGE
    UNIMPORTANT = SYSTEM_UNIMPORTANT
    INFO = SYSTEM_INFO
    DEBUG = SYSTEM_DEBUG
    
    def __init__(self,ignore_level=None,log_file='stderr',
                 log_id=None,log_level=None):

        """ Setup a log object which writes its output to log_file
            using the given id marker. 

            The ignore_level indicates the level at which logging
            activities are ignored. A low value causes less important
            notices like SYSTEM_INFOs to be ignored. If it is set to
            0, no logging is done. 

            Alternatively, log_level may be specified giving the last
            level to log (all higher levels are ignored).

            The log id can be set to separate multiple
            threads/processes in the log. It defaults to the process
            id.

        """
        if ignore_level is not None:
            self.ignore_level = ignore_level
        elif log_level is not None:
            self.ignore_level = log_level + 1
        self.format_log_id(log_id)
        self.log_file = log_file
        self.orig_stderr = _sys.stderr
        self.orig_stdout = _sys.stdout

    def format_log_id(self, log_id=None):

        if not log_id:
            self.log_id = 'p%i' % _os.getpid()
        else:
            self.log_id = '%s:%i' % (log_id, _os.getpid())

    def redirect_stderr(self):

        """ Redirects stderr output to the log mechanism.
        """
        _sys.stderr = self

    def redirect_stdout(self):

        """ Redirects stdout output to the log mechanism.
        """
        _sys.stdout = self

    def disable_redirects(self):

        """ Reset stderr and stdout to their values when the log object
            was created.
        """
        _sys.stderr = self.orig_stderr
        _sys.stdout = self.orig_stdout

    def open(self,flags='a'):

        """ Open the log file. 

            This is done on-the-fly in case a log request is
            encountered and the file is still closed.  This interface
            also supports redirecting messages to stderr or stdout by
            using 'stderr' or 'stdout' as log file name.  If the log
            file cannot be opened, stderr is used instead.

        """
        filename = self.log_file
        if filename == 'stderr':
            file = self.orig_stderr
        elif filename == 'stdout':
            file = self.orig_stdout
        else:
            try:
                file = open(self.log_file,flags)
            except:
                self.open_log_file = file = self.orig_stderr
                log(SYSTEM_ERROR,
                    'Could not open log file "%s" - using stderr',
                    self.log_file)
        self.open_log_file = file
        return file

    def close(self):

        """ Close the log file.

            Note: The next logging request will automatically open the
            file again.

        """
        if self.open_log_file:
            self.open_log_file.close()
        self.open_log_file = None

    def setup(self,ignore_level=None,log_file=None,log_id=None,
              log_level=None):

        """ Change the setup after creation
            - use keywords to change only certain aspects, e.g.
              setup(log_file = 'my.log')
        """
        if ignore_level is not None:
            self.ignore_level = ignore_level
        elif log_level is not None:
            self.ignore_level = log_level + 1
        if log_file is not None:
            self.log_file = log_file
            self.open_log_file = None
        if log_id is not None:
            self.format_log_id(log_id)

    def __call__(self,errno,*args):

        """ Default logging mechanism: errno contains the error id and
            the remaining arguments get interpreted as:
            arg0 % (arg1,arg2,...)
        """
        if errno >= self.ignore_level: return
        if len(args) > 1:
            s = args[0] % args[1:]
        elif len(args) == 1:
            s = args[0]
        else:
            s = ''
        f = self.open_log_file or self.open()
        f.write("%04i %s [%s] %s\n" %
                (errno,_DateTime.now(),self.log_id,s))
        f.flush()

    def traceback(self,errno,*args):

        """ Write a traceback of the last exception to the log file.

        """
        if errno >= self.ignore_level: return
        if len(args) > 1:
            s = args[0] % args[1:]
        elif len(args) == 1:
            s = args[0]
        else:
            s = ''
        f = self.open_log_file or self.open()
        f.write("%04i %s [%s] %s...\n" %
                (errno,_DateTime.now(),self.log_id,s))

        # Print traceback and locals to file
        f.write(79*'-'+ '\n')
        _traceback.print_exc(_TRACEBACK_LIMIT,f)
        print_tb_locals(_sys.exc_info()[2], f,
                        title='### Dump of local variables (omitting private ones):\n',
                        filter=filter_private_attributes)
        f.write(79*'-'+'\n')
        f.flush()

    def modules(self,errno,*args):

        """ Write a list of all currently loaded modules to the log file.
        """
        if errno >= self.ignore_level: return
        if len(args) > 1:
            s = args[0] % args[1:]
        elif len(args) == 1:
            s = args[0]
        else:
            s = ''
        f = self.open_log_file or self.open()
        f.write("%04i %s [%s] %s...\n" %
                (errno,_DateTime.now(),self.log_id,s))
        
        # Print module listing
        f.write(79*'-'+ '\n')
        print_loaded_modules(f)
        f.write(79*'-'+'\n')
        f.flush()

    def stack(self,errno,*args,**kws):

        """ Write a stack dump to the log file.

            A keyword argument "levels" can be given to set the number
            of stack frames to show. It defaults to 3. Higher values
            can result in huge amounts of data to be written to the
            log file !

            If a keyword argument "locals" is given and false, no
            locals are printed in the dump. It defaults to true.

        """
        if errno >= self.ignore_level: return
        if len(args) > 1:
            s = args[0] % args[1:]
        elif len(args) == 1:
            s = args[0]
        else:
            s = ''
        levels = kws.get('levels', 3)
        locals = kws.get('locals', 1)
        f = self.open_log_file or self.open()
        f.write("%04i %s [%s] %s...\n" %
                (errno,_DateTime.now(),self.log_id,s))
        
        # Print stack and locals to file
        f.write(79*'-'+ '\n')
        print_stack(f, levels=levels, offset=1, locals=locals)
        f.write(79*'-'+'\n')
        f.flush()

    def verbatim(self,errno,*args):

        """ Write args[0] % args[1:] to the log file. This method
            is meant for multi line messages, since the logged text
            is not manipulated in any way.
        """
        if errno >= self.ignore_level: return
        if len(args) > 1:
            s = args[0] % args[1:]
        elif len(args) == 1:
            s = args[0]
        else:
            s = ''
        f = self.open_log_file or self.open()
        f.write("%04i %s [%s]\n" % (errno,_DateTime.now(),self.log_id))
        f.write(79*'-'+ '\n%s\n' % s.strip() + 79*'-'+'\n')
        f.flush()

    def call(self,errno,level=1):

        """ Log a function call.

            level indicates how far up the calling stack to look for
            the call information. Default is one level meaning the
            directly calling function.

            Note: Needs mx.Tools.func_call() API.
            
        """
        if errno >= self.ignore_level: return
        from mx.Tools import func_call
        f = self.open_log_file or self.open()
        f.write("%04i %s [%s] %s\n" %
                (errno,_DateTime.now(),self.log_id, func_call(level+1)))
        f.flush()

    def error(self,errno,errorname,*args):

        """ Write errorname, args[0] % args[1:] to the log file.

            This method is meant for multi line error messages that
            hold additional data referring to the error.
            
        """
        if errno >= self.ignore_level: return
        if len(args) > 1:
            s = args[0] % args[1:]
        elif len(args) == 1:
            s = args[0]
        else:
            s = ''
        f = self.open_log_file or self.open()
        f.write("%04i %s [%s] %s\n" %
                (errno,_DateTime.now(),self.log_id, errorname))
        f.write(79*'-'+ '\n%s\n' % str(s).strip() + 79*'-'+'\n')
        f.flush()

    # Alias
    long = error

    def text(self,errno,errorname,*args):

        """ Write errorname and a line numbered version of args[0] %
            args[1:] to the log file.

            This method is meant for multi line error messages that
            hold additional data referring to the error.

            Needs mx.TextTools to be installed.
            
        """
        if errno >= self.ignore_level: return
        if len(args) > 1:
            s = args[0] % args[1:]
        elif len(args) == 1:
            s = args[0]
        else:
            s = ''
        from mx.TextTools import splitlines
        l = splitlines(s)
        for i in range(len(l)):
            l[i] = '%04i %s' % (i+1,l[i])
        f = self.open_log_file or self.open()
        f.write("%04i %s [%s] %s\n" %
                (errno,_DateTime.now(),self.log_id, errorname))
        f.write(79*'-'+ '\n%s\n' % '\n'.join(l) + 79*'-'+'\n')
        f.flush()

    def object(self,errno,objname,object,levels=2):

        """ Write objname plus a dump of object to the log file.

            levels indicates how many levels of the object hierarchy
            should be written.
            
        """

        if errno >= self.ignore_level: return
        f = self.open_log_file or self.open()
        f.write("%04i %s [%s] %s\n" %
                (errno, _DateTime.now(), self.log_id, objname))
        f.write(79*'-'+'\n')
        print_obj(object,f,levels=levels)
        f.write(79*'-'+'\n')
        f.flush()
        
    def write(self,data):

        """ File interface, so that the object can be used as
            sys.stderr replacement.

            Errorcode 0 is used. The output is line buffered.
            
        """
        if self.ignore_level == 0: return
        self.line_buffer = buffer = self.line_buffer + data
        if '\n' not in buffer:
            return

        # Write out all complete lines stored in line_buffer
        lines = buffer.split('\n')
        if buffer[-1:] != '\n':
            # last line is incomplete
            self.line_buffer = lines[-1]
            lines = lines[:-1]
        else:
            self.line_buffer = ''

        f = self.open_log_file or self.open()
        now = _DateTime.now()
        id = self.log_id
        fwrite = f.write
        for line in lines:
            if not line.strip():
                # avoid emtpy lines
                continue
            fwrite("%04i %s [%s] %s\n" % (0,now,id,line))
        f.flush()

    def flush(self):

        """ Flush the underlying log stream.
        """
        if self.open_log_file:
            self.open_log_file.flush()

    def exception(self,errno,*args):

        """ Log a real exception in the log file.
        """
        if errno >= self.ignore_level: return
        exc_class,exc_inst,exc_traceback = _sys.exc_info()
        if exc_class is None:
            return
        if len(args) > 1:
            s = args[0] % args[1:]
        elif len(args) == 1:
            s = args[0]
        else:
            s = ''

        # go down the traceback
        tb = exc_traceback
        while tb.tb_next is not None:
            tb = tb.tb_next
        frame = tb.tb_frame

        f = self.open_log_file or self.open()
        f.write("%04i %s [%s] %s (at '%s':%i - %s:'%s')\n" % 
                (errno,_DateTime.now(),self.log_id,s,
                 frame.f_code.co_filename,
                 _traceback.tb_lineno(tb),
                 exc_class,exc_inst))
        f.flush()

        # Cleanup to avoid circ. refs
        del exc_class,exc_inst,exc_traceback

### Variants of the main Log class #####################################

### ListLog logs to a list instead of a file

class ListFile:

    """ Wraps a list using a file-like interface.

        Just enough to work as log file for mxLog.

    """

    # Log list
    data = None

    def __init__(self, data, *args):
        self.data = data

    def write(self, info):
        self.data.append(info)

    def flush(self):
        pass

    def close(self):
        pass

class ListLog(Log):

    """ mxLog variant that logs to a list instead of a file.

    """

    def __init__(self, log_list, *args, **kws):

        self.open_log_file = ListFile(log_list)
        Log.__init__(self, *args, **kws)

    def close(self):
        pass

### ThreadLog includes the thread id in the log_id

try:
    import threading as _threading

except (ImportError, NameError):
    pass

else:

    class ThreadLog(Log, object):

        """ mxLog variant logging both the PID and thread name of the
            caller

            It can be used as a common logging facility for multiple
            threads, since log output from each thread can be identified
            by thread name.

        """
        # Custom log_id set via e.g. .setup()
        _custom_log_id = ''

        def format_log_id(self, log_id=None):

            if log_id is not None:
                self._custom_log_id = log_id

        def _get_log_id(self):

            if self._custom_log_id:
                return '%s:%i/%s' % (
                    self._custom_log_id,
                    _os.getpid(),
                    _threading.currentThread().getName())
            else:
                return 'p%i/%s' % (
                    _os.getpid(),
                    _threading.currentThread().getName())

        def _set_log_id(self, value):

            self._custom_log_id = value

        log_id = property(_get_log_id, _set_log_id)

### LogNothing disabled logging completely

class LogNothing(Log):

    # Don't log anything
    ignore_level = SYSTEM_LOG_NOTHING

    def setup(self, *args, **kws):
        Log.setup(*args, **kws)
        self.ignore_level = SYSTEM_LOG_NOTHING

    def open(self, flags='a'):
        self.open_log_file = None

    def close(self):
        pass
    
###

# Create a main log object
if __debug__ and Tools.debugging():
    log = Log(SYSTEM_LOG_EVERYTHING)
else:
    log = Log(SYSTEM_INFO + 1)



import mx.Tools.NewBuiltins

class AcquisitionMixin:

    """ Mixin class allowing implicit acquisition of attributes from
        objects up a containment hierachy.

        To enable this feature, an object 'b' contained in another
        object 'a' must set the instance variable b.baseobj to
        a. The search is done recursively until either an exception
        is raised by the queried object or the object is found.

        Attributes that strt with an underscore ('_') can not
        be acquired. Be careful when acquiring methods: the method
        will be executed in the baseobjects context, not the one
        it is acquired into.

        See the documentation of acquire() for details.

    """
    baseobj = None              # Base object ("containing" this object)

    # Implicit acquisition
    __getattr__ = acquire       # Use the (new) builtin acquire() to implement
                                # the acquisition mechanism

    def __init__(self,*args,**kw):

        """ Create the object and call startup(*args,**kw)
            - if you give a keyword 'baseobj' then its parameter
              is used to set the baseobj for acquisition; the keyword
              is then deleted from kw
        """
        # Set baseobj for acquisition to work properly
        baseobj = kw.get('baseobj',None)
        if baseobj is not None:
            self.baseobj = baseobj
            del kw['baseobj']
        # call startup()
        apply(self.startup,args,kw)

    def startup(self,*args,**kw):

        """ Override this to init the instance.
        """
        pass

class PythonAcquisitionMixin(AcquisitionMixin):

    def __getattr__(self,name,

                    getattr=getattr,AttributeError=AttributeError):

        if name[0] != '_':
            return getattr(self.baseobj,name)
        raise AttributeError,name

    # Localize these here for speed
    baseobj = AcquisitionMixin.baseobj
    __init__ = AcquisitionMixin.__init__
    startup = AcquisitionMixin.startup

def _test():

    class C(AcquisitionMixin):
        a = 2
        x = 9

    class D(AcquisitionMixin):
        b = 3

    class E(AcquisitionMixin):
        c = 4

    # Setup acquisition chain
    c = C()
    d = D(baseobj=c)
    e = E(baseobj=d)

    # Give implicit acquisition a go...
    print c.a, d.a, e.a
    print d.b, e.b
    print e.c

    d.a = 5
    print c.a, d.a, e.a

    c.b = 1
    print c.b, d.b, e.b
    print
    
    # Performance comparism:
    print 'Performance:'
    print
    import time
    l = range(100000)

    class C(AcquisitionMixin):
        x = 9

    class D(AcquisitionMixin):
        y = 8

    class E(AcquisitionMixin):
        z = 7

    # Setup acquisition chain
    c = C()
    d = D(baseobj=c)
    e = E(baseobj=d)

    t = time.time()
    for i in l:
        e.x
        e.y
        e.z
    print 'AcquisitionMixin:',time.time() - t,'seconds'

    class C(PythonAcquisitionMixin):
        x = 9

    class D(PythonAcquisitionMixin):
        y = 8

    class E(PythonAcquisitionMixin):
        z = 7

    # Setup acquisition chain
    c = C()
    d = D(baseobj=c)
    e = E(baseobj=d)

    t = time.time()
    for i in l:
        e.x
        e.y
        e.z
    print 'PythonAcquisitionMixin:',time.time() - t,'seconds'

if __name__ == '__main__':
    _test()




""" Cache - Generic cache implementation

    Copyright (c) 1998-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
    See the documentation for further information on copyrights,
    or contact the author. All Rights Reserved.
"""
import sys
from mx import Tools

# Weight table
MAXLOCALITY = 256
_weights = [1] * MAXLOCALITY

# maximal history size (must somewhat larger than MAXLOCALITY)
HISTORYLIMIT = 2 * MAXLOCALITY

# Cut goal that has to be reached (self.max_cachesize * float(DENOM / NOM))
DENOM = 3
NOM = 4

# Print debugging info ?
_debug = 0

# Init globals
class _modinit:
    import math
    l = Tools.frange(0,1,MAXLOCALITY)
    for i,factor in Tools.irange(l):
        _weights[i] = int((math.exp(factor) - 1.0) * 8192)
        if _debug:
            print i,'. weight =',_weights[i]
del _modinit

### Classes

class NotCached:

    """ A singleton that can be used in conjunction with the .get()
        method of Cache. It behaves like None.
    """
    def __nonzero__(self):
        return 0
    __len__ = __nonzero__

NotCached = NotCached()

###

class Cache:

    """ Tunable cache implementation

        The following parameters influence the cache behaviour:
        - max_cachesize:        the cache will be cut smaller when this limit
                                is reached
        - max_elementsize:      elements with size greate than this limit
                                will not be cached
        - locality:             these many elements will be looked at in
                                the hit statistics to determine how important
                                a specific entry is
    """
    data = None                 # Dict. of id:value cache entries
    put_history = None          # Reverse list of id puts; last is most
                                # recent access; contains id for each access
    get_history = None          # Reverse list of id gets; last is most
                                # recent access; contains id for each access

    def __init__(self,max_cachesize=200,
                      max_elementsize=4096,
                      locality=50):

        self.max_cachesize = max_cachesize
        self.max_elementsize = max_elementsize
        if locality > MAXLOCALITY:
            raise ValueError,'locality must be <= %i' % MAXLOCALITY
        self.locality = locality

        self.cachesize = 0
        self.cuts = 0
        self.misses = 0
        self.hits = 0

        self.data = {}
        self.put_history = []
        self.get_history = []

    def cut(self,

            NOM=NOM,DENOM=DENOM):

        """ Force a cut of the cache's contents.

            This will make room for at least one new entry.
        """
        if _debug:
            print '  Cutting down cache size...'
        cachesize = self.cachesize
            
        # Cut the cache down to the entries in recent get history
        newdata = {}
        known_key = newdata.has_key
        data = self.data
        for id in self.get_history[-self.locality:]:
            if known_key(id):
                continue
            try:
                newdata[id] = data[id]
            except KeyError:
                pass

        cachesize = len(newdata)
        if _debug:
            print '   Size after cut to recent history:',cachesize

        # Check
        if cachesize * NOM >= self.max_cachesize * DENOM:

            # Calculate weights
            d = {}
            weights = _weights
            d_get = d.get
            for i,id in Tools.irange(self.get_history[-self.locality:]):
                if not known_key(id):
                    continue
                d[id] = d_get(id,0) + weights[i]

            # Delete all entries left from median
            ranking = Tools.sortedby(d.items(),1)
            if _debug:
                print '   Ranking:',ranking
            for id,weight in ranking[:len(d)/2]:
                if _debug:
                    print '   Deleting',id,'with weight =',weight
                del newdata[id]

            # Check
            cachesize = len(newdata)

            if cachesize * NOM >= self.max_cachesize * DENOM:
                # Ok, so the smart way didn't work...
                if _debug:
                    print '   Did not work, going the hard way...'
                newdata.clear()
                cachesize = 0

        self.data = newdata
        self.cachesize = cachesize
        self.cuts = self.cuts + 1

    def clear(self):

        """ Clear the cache.
        """
        self.cachesize = 0
        self.data = {}
        self.history = []

    def get(self,id,default=NotCached,

            HISTORYLIMIT=HISTORYLIMIT):

        """ Get a value from the cache or return default if it is
            not cached.
        """
        item = self.data.get(id,None)
        if item is None:
            self.misses = self.misses + 1
            return default

        # Add "hit"
        self.get_history.append(id)
        if len(self.get_history) > HISTORYLIMIT:
            del self.get_history[-self.locality:]
        self.hits = self.hits + 1

        return item

    def put(self,id,value,sizeof=Tools.sizeof,

            HISTORYLIMIT=HISTORYLIMIT):

        """ Add a value to the cache or update an existing one.
        """
        size = sizeof(value)
        if size > self.max_elementsize:
            return

        # Adding a new entry: make sure there is room
        if not self.data.has_key(id):
            if _debug:
                print '  Adding',id
            self.cachesize = cachesize = self.cachesize + 1
            if cachesize > self.max_cachesize:
                self.cut()

        self.data[id] = value
        self.put_history.append(id)
        if len(self.put_history) > HISTORYLIMIT:
            del self.put_history[-self.locality:]


    # Aliases
    add = put
    update = put

    def delete(self,id):

        """ Delete an entry from the cache.

            It is not an error, if the entry is not currently in the cache.
        """
        try:
            del self.data[id]
        except KeyError:
            pass

    # Aliases
    remove = delete

###

if __name__ == '__main__':
    c = Cache(10,100,locality=17)
    i = 1
    while 1:
        print 'Round',i,c.cachesize,c.cuts,c.misses,c.data.keys(); i = i + 1
        c.put(1,2)
        c.get(1)
        c.get(1)
        c.put(5,2)
        c.get(5)
        c.put(2,2)
        c.put(3,2)
        c.get(2)
        c.get(3)
        c.put(2,3)
        c.put(3,'x'*200)
        c.get(2)
        c.get(3)
        c.get(2)
        c.put(4,2)
        c.get(4,2)
        c.get(4,2)
        c.get(4,2)
        # Add noise
        c.put(i,'x')
        c.put(i * 2,'x')
        c.put(i * 3,'x')
        c.get(i)




""" PackageTools - A set of tools to aid working with packages.

    Copyright (c) 1998-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
    Copyright (c) 2000-2013, eGenix.com Software GmbH; mailto:info@egenix.com
    See the documentation for further information on copyrights,
    or contact the author. All Rights Reserved.
"""
__version__ = '0.4.0'

import os,types,sys,re,imp,__builtin__
import mx.Tools.NewBuiltins

# RE to identify Python modules
suffixes = projection(imp.get_suffixes(),0)
module_name = re.compile('(.*)(' + '|'.join(suffixes) + ')$')
initmodule_name = re.compile('__init__(' + '|'.join(suffixes) + ')$')
initmodule_names = []
for suffix in suffixes:
    initmodule_names.append('__init__' + suffix)

def find_packages(dir=os.curdir, files_only=0, recursive=0, ignore_modules=0,
                  pkgbasename='', pkgdict=None,
                  
                  isdir=os.path.isdir,exists=os.path.exists,
                  isfile=os.path.isfile,join=os.path.join,listdir=os.listdir,
                  module_name=module_name,initmodule_name=initmodule_name):

    """ Return a list of package names found in dir.

        Packages are Python modules and subdirectories that provide an
        __init__ module.  The .py extension is removed from the
        files. The __init__ modules are not considered being seperate
        packages.

        If files_only is true, only Python files are included in the
        search (subdirectories are *not* taken into account). If
        ignore_modules is true (default is false), modules are
        ignored. If recursive is true the search recurses into package
        directories.

        pkgbasename and pkgdict are only used during recursion.
        
    """
    l = listdir(dir)
    if pkgdict is None:
        pkgdict = {}
    if files_only:
        for filename in l:
            m = module_name.match(filename)
            if m is not None and \
               m.group(1) != '__init__':
                pkgdict[pkgbasename + m.group(1)] = 1
    else:
        for filename in l:
            path = join(dir, filename)
            if isdir(path):
                # Check for __init__ module(s)
                for name in initmodule_names:
                    if isfile(join(path, name)):
                        pkgname = pkgbasename + filename
                        pkgdict[pkgname] = 1
                        if recursive:
                            find_packages(path,
                                          recursive=1,
                                          pkgbasename=pkgname + '.',
                                          pkgdict=pkgdict)
                        break
            elif not ignore_modules:
                m = module_name.match(filename)
                if m is not None and \
                   m.group(1) != '__init__':
                    pkgdict[pkgbasename + m.group(1)] = 1
    return pkgdict.keys()

def find_subpackages(package, recursive=0,

                     splitpath=os.path.split):

    """ Assuming that package points to a loaded package module, this
        function tries to identify all subpackages of that package.

        Subpackages are all Python files included in the same
        directory as the module plus all subdirectories having an
        __init__.py file.  The modules name is prepended to all
        subpackage names.

        The module location is found by looking at the __file__
        attribute that non-builtin modules define. The function uses
        the __all__ attribute from the package __init__ module if
        available.

        If recursive is true (default is false), then subpackages of
        subpackages are recursively also included in the search.
        
    """
    if not recursive:
        # Try the __all__ attribute...
        try:
            subpackages = list(package.__all__)
        except (ImportError, AttributeError):
            # Did not work, then let's try to find the subpackages by looking
            # at the directory where package lives...
            subpackages = find_packages(package.__path__[0], recursive=recursive)
    else:
        # XXX Recursive search does not support the __all__ attribute
        subpackages = find_packages(package.__path__[0], recursive=recursive)
    basename = package.__name__ + '.'
    for i,name in irange(subpackages):
        subpackages[i] = basename + name
    return subpackages

def _thismodule(upcount=1,

                exc_info=sys.exc_info,trange=trange):

    """ Returns the module object that the callee is calling from.

        upcount can be given to indicate how far up the execution
        stack the function is supposed to look (1 == direct callee, 2
        == callee of callee, etc.).

    """
    try:
        1/0
    except:
        frame = exc_info()[2].tb_frame
        for i in trange(upcount):
            frame = frame.f_back
    name = frame.f_globals['__name__']
    del frame
    return sys.modules[name]

def _module_loader(name, locals, globals, sysmods, errors='strict',
                   importer=__import__, reloader=reload, from_list=['*']):

    """ Internal API for loading a module
    """
    if not sysmods.has_key(name):
        is_new = 1
    else:
        is_new = 0
    try:
        mod = importer(name, locals, globals, from_list)
        if reload and not is_new:
            mod = reloader(mod)
    except KeyboardInterrupt:
        # Pass through; SystemExit will be handled by the error handler
        raise
    except Exception, why:
        if errors == 'ignore':
            pass
        elif errors == 'strict':
            raise
        elif callable(errors):
            errors(name, sys.exc_info()[0], sys.exc_info()[1])
        else:
            raise ValueError,'unknown errors value'
    else:
        return mod
    return None

def import_modules(modnames,module=None,errors='strict',reload=0,

                   thismodule=_thismodule):

    """ Import all modules given in modnames into module.

        module defaults to the caller's module. modnames may contain
        dotted package names.

        If errors is 'strict' (default), then ImportErrors and
        SyntaxErrors are raised. If set to 'ignore', they are silently
        ignored. If errors is a callable object, then it is called
        with arguments (modname, errorclass, errorvalue). If the
        handler returns, processing continues.

        If reload is true (default is false), all already modules
        among the list will be forced to reload.
        
    """
    if module is None:
        module = _thismodule(2)
    locals = module.__dict__
    sysmods = sys.modules
    for name in modnames:
        mod = _module_loader(name, locals, locals, sysmods, errors=errors)
        if mod is not None:
            locals[name] = mod
    
def load_modules(modnames,locals=None,globals=None,errors='strict',reload=0):

    """ Imports all modules in modnames using the given namespaces and returns
        list of corresponding module objects.

        If errors is 'strict' (default), then ImportErrors and
        SyntaxErrors are raised. If set to 'ignore', they are silently
        ignored. If errors is a callable object, then it is called
        with arguments (modname, errorclass, errorvalue). If the
        handler returns, processing continues.

        If reload is true (default is false), all already modules
        among the list will be forced to reload.
        
    """
    modules = []
    append = modules.append
    sysmods = sys.modules
    for name in modnames:
        mod = _module_loader(name, locals, globals, sysmods, errors=errors)
        if mod is not None:
            append(mod)
    return modules
    
def import_subpackages(module, reload=0, recursive=0,

                       import_modules=import_modules,
                       find_subpackages=find_subpackages):

    """ Does a subpackages scan using find_subpackages(module) and then
        imports all submodules found into module.

        The module location is found by looking at the __file__
        attribute that non-builtin modules define. The function uses
        the __all__ attribute from the package __init__ module if
        available.

        If reload is true (default is false), all already modules
        among the list will be forced to reload.
        
    """
    import_modules(find_subpackages(module, recursive=recursive),
                   module, reload=reload)

def load_subpackages(module, locals=None, globals=None, errors='strict', reload=0,
                     recursive=0,
                     
                     load_modules=load_modules,
                     find_subpackages=find_subpackages):

    """ Same as import_subpackages but with load_modules
        functionality, i.e. imports the modules and also returns a list of
        module objects.

        If errors is 'strict' (default), then ImportErrors are
        raised. If set to 'ignore', they are silently ignored.

        If reload is true (default is false), all already modules
        among the list will be forced to reload.
        
    """
    return load_modules(find_subpackages(module, recursive=recursive),
                        locals, globals,
                        errors=errors, reload=reload)

def modules(names,

            extract=extract):

    """ Converts a list of module names into a list of module objects.
    
        The modules must already be loaded.

    """
    return extract(sys.modules, names)

def package_modules(pkgname):

    """ Returns a list of all modules belonging to the package with the
        given name.

        The package must already be loaded. Only the currently
        registered modules are included in the list.
        
    """
    match = pkgname + '.'
    match_len = len(match)
    mods = [sys.modules[pkgname]]
    for k,v in sys.modules.items():
        if k[:match_len] == match and v is not None:
            mods.append(v)
    return mods

def find_classes(mods,baseclass=None,annotated=0,

                 ClassType=types.ClassType,issubclass=issubclass):

    """ Find all subclasses of baseclass or simply all classes (if baseclass
        is None) defined by the module objects in list mods.

        If annotated is true the returned list will contain tuples
        (module_object,name,class_object) for each class found where
        module_object is the module where the class is defined.
        
    """
    classes = []
    for mod in mods:
        for name,obj in mod.__dict__.items():
            if type(obj) is ClassType:
                if baseclass and not issubclass(obj,baseclass):
                    continue
                if annotated:
                    classes.append((mod, name, obj))
                else:
                    classes.append(obj)
    return classes

def find_instances(mods,baseclass,annotated=0,

                   InstanceType=types.InstanceType,issubclass=issubclass):

    """ Find all instances of baseclass defined by the module objects
        in list mods.

        If annotated is true the returned list will contain tuples
        (module_object,name,instances_object) for each instances found where
        module_object is the module where the instances is defined.
        
    """
    instances = []
    for mod in mods:
        for name,obj in mod.__dict__.items():
            if isinstance(obj,baseclass):
                if annotated:
                    instances.append((mod,name,obj))
                else:
                    instances.append(obj)
    return instances
